// 5648-C84, 5648-C83, (C) Copyright IBM Corporation, 1997, 2000
// All rights reserved. Licensed Materials Property of IBM
// Note to US Government users: Documentation related to restricted rights
// Use, duplication or disclosure is subject to restrictions set forth in GSA ADP Schedule with IBM Corp.
// This page may contain other proprietary notices and copyright information, the terms of which must be observed and followed.
//
// This program may be used, executed, copied, modified and distributed
// without royalty for the purpose of developing, using,
// marketing, or distributing.

/*
 * This is the core of the SiteOutliner function.  It can be invoked from
 * a command line or via a servlet front-end (SiteOutlinerServlet).
 *
 * <p>Refer to SiteOutliner's help file (about.html) for usage information.
 */

package WebSphereSamples.SiteOutliner;

import java.io.*;
import java.net.*;
import java.util.*;
import java.text.*;
import org.w3c.dom.*;
import com.ibm.xml.parser.*;

/**
  * Example SiteOutliner-- 
  * This class demonstrates document generation functions of XML Document Structure Services.
  */
public class SiteOutliner extends Object {

   // Class variables
   static String      s_rootDir         = null;                 // Root directory to recursively scan from
   static int         s_depth           = 10;                   // Depth to recursively scan the server's document tree
   static String      s_documentBaseURL = null;                 // Document base URL to append scanned file path names to
   static String[]    s_excludeList     = {};                   // Directories or file to exclude from consideration (e.g. /backup, /archive)
   static String[]    s_fileTypes       = {".html",".htm",".HTML",".HTM"}; // File types to consider
   static String      s_encoding        = null;                 // Input encoding (Java encoding)
   static String      s_outEncoding     = null;                 // Output encoding (MIME charset encoding)
   static String      s_cdfFileName     = "SiteOutliner";       // .CDF file name
   static boolean     s_debug           = false;                // Debug mode?

   // Instance variables
   private File       rootDir           = null;
   private int        depth             = 10;
   private String     documentBaseURL   = null;
   private String[]   excludeList       = null;
   private String[]   fileTypes         = null;
   private String     encoding          = null;
   private String     outEncoding       = null;

   // XML Document Structure Services variables
   private Document   doc               = null;                 // Root document object                 



   /** 
     * Entry point for application invocation.  
     *
     * @param args Invocation options and parameters (see usage method for details).
     * @exception ClassNotFoundException When XML Document Structure Services not found.
     * @exception InstantiationException When XML Document Structure Services not found.
     * @exception IllegalAccessException When XML Document Structure Services not found.
     * @see #usage
     */
   public static void main(String[] args) 
   throws ClassNotFoundException, InstantiationException, IllegalAccessException {

      // Parse command line parameters and check for required parameters.
      parseArgs(args);  

      // Echo what we think the key processing parameters are.
      System.err.println("Scan starting from "+s_rootDir+", document base prefix="+s_documentBaseURL);
      System.err.print("  Looking for files that end with ");
      int i;
      for (i=0; i<s_fileTypes.length; i++)
         System.err.print(s_fileTypes[i]+";");
      System.err.println();
      System.err.print("  Except for ");
      for (i=0; i<s_excludeList.length; i++)
         System.err.print(s_excludeList[i]+";");
      System.err.println();

      // Create an instance of SiteOutliner and set its retained data.
      SiteOutliner aSiteOutliner  = new SiteOutliner(s_rootDir,s_documentBaseURL);
      aSiteOutliner.depth         = s_depth;
      aSiteOutliner.excludeList   = s_excludeList;
      aSiteOutliner.fileTypes     = s_fileTypes;
      aSiteOutliner.encoding      = s_encoding;
      aSiteOutliner.outEncoding   = s_outEncoding;

      // Generate the XML document and print it out.
      aSiteOutliner.buildDocument();
      aSiteOutliner.printDocument();
   }

   /** 
     * Constructor. This method will also provide semantic checking of
     * <var>rootDir</var> and <var>documentBaseURL</var>.
     *
     * @param rootDir          Root directory to recursively scan from.
     * @param documentBaseURL  Document base URL to append scanned file path names to.
     */
   public SiteOutliner(String rootDir,String documentBaseURL) {

      // Check if documentBaseURL is valid
      this.documentBaseURL = documentBaseURL;
      try {
         URL url = new URL(this.documentBaseURL);
      } catch (MalformedURLException e) {
         fatalError("Specified document base URL (documentBaseURL) parameter can not be formed into a URL. "+e.getMessage());
      }
      try {
         // Check if rootDir is valid   
         this.rootDir = new File(rootDir);
         if (!this.rootDir.isDirectory()) {
            fatalError(this.rootDir+" is not a directory on this server.");
         }
         this.rootDir = new File(this.rootDir.getAbsolutePath());
      } catch (Throwable e) {
         fatalError("Specified root directory (rootDir) parameter is not valid.  "+e.getMessage());
      }
   }

   /* ----------------------------- Private Methods ------------------------- */

   /**
     * Parse command line parameters, and check for required parameters.
     * 
     * @param args Command line parameters.
     * @see #usage
     */
   private static void parseArgs(String[] args) {
      int n = 0;
      for (int i=0; i<args.length; i++) {   // For each parameter..
         if (args[i].charAt(0)=='-') {       // Options start with -
            if (args[i].equals("-debug")) {
               s_debug = true;
            } else if (args[i].equals("-filetype")) {
               i++; 
               if (i>=args.length)
                  fatalError("Value(s) of the file type (-filetype) option is missing.  ");
               StringTokenizer tn = new StringTokenizer(args[i],";",false);
               s_fileTypes = new String[tn.countTokens()];
               for (int j=0; tn.hasMoreTokens(); j++) {
                  String s = (String)tn.nextToken();
                  s_fileTypes[j] = s;
               }
            } else if (args[i].equals("-exclude")) {
               i++; 
               if (i>=args.length)
                  fatalError("Value(s) of the exclude (-exclude) option is missing.  ");
               StringTokenizer tn = new StringTokenizer(args[i],";",false);
               s_excludeList = new String[tn.countTokens()];
               for (int j=0; tn.hasMoreTokens(); j++) {
                  String s = (String)tn.nextToken();
                  s_excludeList[j] = s;
               }
            } else if (args[i].equals("-encoding")) {
               i++; 
               if (i>=args.length)
                  fatalError("Value of the input encoding (-encoding) option is missing.  ");
               s_encoding = args[i];
            } else if (args[i].equals("-outencoding")) {
               i++; 
               if (i>=args.length)
                  fatalError("Value of the output encoding (-outencoding) option is missing.  ");
               s_outEncoding = args[i];
            } else if (args[i].equals("-depth")) {
               i++; 
               if (i>=args.length)
                  fatalError("Value of the depth (-depth) option is missing.  ");
               s_depth = java.lang.Integer.parseInt(args[i]);
            } else if (args[i].equals("-help")) {
               usage(0);
            } else { // Unknown parameter
               fatalError("Unknown parameter encountered: "+args[i]);
            }
         }  // Option seen 
         else if (n==0) {  // 1st positional parmeter is rootDir
            s_rootDir = args[i];
            n++;
         } else if (n==1) {  // 2nd positional parameters is documentBaseURL
            s_documentBaseURL = args[i];
            n++;
         }
      } // For each parameter..
      if (s_rootDir==null)         // rootDir is required 
         fatalError("Root directory (rootDir) parameter must be specified.");
      if (s_documentBaseURL==null) // documentBaseURL is required
         fatalError("Document base URL (documentBaseURL) parameter must be specified.");
   }

   /**
     * Display error message and correct command line usage info...then exit
     * with a non-zero code.
     * 
     * @param errorMessage Message to be displayed with the prefix &quot;Error:&quot;.
     */
   private static void fatalError(String errorMessage) {
      System.err.println("Error: "+errorMessage);
      System.err.println("");
      usage(-1);
   }

   /**
     * Display SiteOutliner invocation format and options.
     * 
     * @param exitCode   Value to pass on System.exit after printing usage.
     */
   private static void usage(int exitCode) {
      System.err.println("Usage: java SiteOutliner [option(s)] rootDir documentBaseURL");
      System.err.println("");
      System.err.println("Required parameters:");
      System.err.println("   rootDir: Server root directory from which to recursively scan");
      System.err.println("   documentBaseURL: URL prefix which represents rootDir");
      System.err.println("Options:");
      System.err.println("   -depth <depth>: Depth to recursively scan down rootDir, default=10");
      System.err.println("   -filetype <type1;type2;...>: File types to consider, default=.htm;.html;.HTM;.HTML");
      System.err.println("   -exclude <name1;name2;...>: File and directory names to exclude from consideration, default=");
      System.err.println("   -encoding <javaEncoding>: Encoding for input, default=default Java encoding");
      System.err.println("   -outencoding <MIMEcharset>: Encoding for output, default=derived from default Java encoding");
      System.err.println("   -debug: Run with debug mode on, default=off");
      System.err.println("   -help: Display this help text");
      System.err.println("");
      System.err.println("Example: java SiteOutliner -exclude archive -encoding JISAutoDetect -outencoding ISO-2022-JP c:/InetPub/wwwroot http://www.ibm.com");
      System.exit(exitCode);
   }

   /** 
     * Creates a site outline (channel) in CDF format from the specified <var>rootDir</var>
     * to the recursive depth specified by the <var>-depth</var> option.
     *
     * <p>A &lt;CHANNEL&gt; element will be created for each non-empty directory in
     * in the hierarchy.
     *
     * <p>An &lt;ITEM&gt; element will be created for each file in a directory.
     *
     * <p>Files and directories are filtered in accordance with the <var>-filetype</var>
     * and <var>-exclude</var> options.
     * @exception ClassNotFoundException When XML Document Structure Services not found.
     * @exception InstantiationException When XML Document Structure Services not found.
     * @exception IllegalAccessException When XML Document Structure Services not found.
     */
   private void buildDocument() 
   throws ClassNotFoundException, InstantiationException, IllegalAccessException {

      // Create an XML document object, and add the root element.
      this.doc = (Document)Class.forName("com.ibm.xml.parser.TXDocument").newInstance();
      Element rootElement = this.doc.createElement("CHANNEL");
      this.doc.appendChild(rootElement);

      // Recursively scan down from the root directory and build sub-channels.
      processDirectory(rootElement,this.rootDir,this.documentBaseURL,this.depth);
   }

   /**
     * Process a directory into the site outline. 
     *
     * <p>See buildDocument for a description of processing at the directory level.
     * 
     * @param channelElement   The DOM Element which defines the (sub-)channel  
     *                         for the directory being processed.
     * @param dir              The current directory being processed; this is a
     *                         derivative of <var>rootDir</var>.
     * @param urlPrefix        The current URL prefix that represents <var>var</var>;
     *                         this a derivative of <var>documentBaseURL</var>.
     * @param curDepth         The depth of the directory hierarchy yet to be traversed.
     * @return                 The number of non-filtered and non-empty directories plus
     *                         the number of non-filtered files.
     * @see #buildDocument
     */
   private int processDirectory(Element channelElement, File dir, 
                                String urlPrefix, int curDepth) {

      int     nDocuments  = 0;
      String  dirTitle    = urlPrefix;
      File    file        = null;
      String  url         = null;

      // Stop recursive decent when we've reached the appropriate depth.
      if (curDepth-- <= 0)
         return 0;

      // "Chat it up a bit" if we're in debug mode.
      if (s_debug)
         System.err.println("Processing "+dir.getAbsolutePath()+"[current depth="+curDepth+"]");

      // Examine directory contents and begin processing...
      try {
         String[] fileNameList = dir.list();
         if (fileNameList!=null) {  // If directory is not empty..

            outer:  for (int i=0; i<fileNameList.length; i++) {  // For each file or dir..

               // If file or directory is on the exclude list, ignore it.
               for (int j=0; j<this.excludeList.length; j++) {
                  if (fileNameList[i].equals(this.excludeList[j]))
                     continue outer;
               }

               // Build relative filespec (to rootDir) and URL (to documentBaseURL).
               file = new File(dir,fileNameList[i]);
               url  = urlPrefix+"/"+fileNameList[i];

               // If content is a subdirectory..
               if (file.isDirectory()) {

                  // Create a new <CHANNEL> DOM element.
                  Element newChannelElement = this.doc.createElement("CHANNEL");

                  // Recurse down to process the directory.
                  int n = processDirectory(newChannelElement, file, url, curDepth);

                  // Add the new sub-channel as a DOM Child Node to channelElement
                  // only if it contains (filtered) content.
                  if (n > 0) {
                     channelElement.appendChild(newChannelElement);
                     nDocuments += n;
                  }
               }

               // If content is a non-filtered file..
               else if (filterFile(file)) {

                  // Create an <ITEM HREF=url> DOM element.
                  Element itemElement = this.doc.createElement("ITEM");
                  itemElement.setAttribute("HREF",url);

                  // Create a <TITLE>item title</TITLE> DOM element, and add it as a 
                  // DOM Child Node to <ITEM>.
                  Element titleElement = this.doc.createElement("TITLE");
                  itemElement.appendChild(titleElement);
                  String itemTitle = getTitle(file);
                  if (itemTitle == null)
                     itemTitle = "No title";
                  titleElement.appendChild(this.doc.createTextNode(itemTitle));

                  // Create an <ABSTRACT> DOM element, and add it as a DOM Child
                  // Node to <ITEM>.
                  itemElement.appendChild(this.doc.createElement("ABSTRACT"));

                  // Add the <ITEM> Node (and its children) as a DOM Child Node
                  // to channelElement.
                  channelElement.appendChild(itemElement);

                  // Bump our count of non-filtered files at this level of recursion.
                  nDocuments++;

                  // index.htm's item title overrides the default directory title.
                  if (fileNameList[i].startsWith("index.htm")) {
                     dirTitle = itemTitle;
                  }
               }  // If a non-filtered file

               // Even if index.htm is filtered, override the default directory title.
               else if (fileNameList[i].startsWith("index.htm")) {
                  dirTitle = getTitle(file);
                  if (dirTitle == null)
                     dirTitle = "No title";
               }

            }  // For each file or dir..
         }  // If directory is not empty..

         // Create a <TITLE>channel title (#)</TITLE> DOM element, and add it as a 
         // DOM Child Node to channelElement.
         Element title = this.doc.createElement("TITLE");
         title.appendChild(this.doc.createTextNode(dirTitle+" ("+nDocuments+")"));
         channelElement.appendChild(title);

         // Add a DOM attribute to channelELement to make it look like <CHANNEL HREF=url>.
         channelElement.setAttribute("HREF",urlPrefix);

      } catch (Throwable e) {
         if (file!=null) {
            System.err.println("Error: processDirectory(): "+e.getMessage()+" File="+file.getAbsolutePath());
         } else {
            System.err.println("Error: processDirectory(): "+e.getMessage()+" Dir="+dir);
         }
      }
      return nDocuments;
   }

   /**
     * Returns the title from the specified <var>file</var> by looking for 
     * the contents of the HTML <title> tag.
     *
     * <p>Note that if <var>file</var> was guaranteed to be well formed XML, it 
     * would be rather straight forward to parse the file and extract the DOM text
     * Node child of the first <title> DOM element we encountered.
     * 
     * @param file File to be examined.  
     * @return     The text from the HTML <title> tag, or <var>null</var> if no tag.
     */
   private String getTitle(File file) {

      String           title = "";
      String           line;
      BufferedReader   aBufferedReader;
      int              i;

      // "Chat it up a bit" if we're in debug mode.
      if (s_debug) {
         System.err.println("Processing file "+file.getAbsolutePath()+" with encoding "+this.encoding);
      }

      // Read file and be sensitive to input encoding if it was specified.
      try {
         if (this.encoding!=null) {
            aBufferedReader = new BufferedReader(new InputStreamReader(new FileInputStream(file),this.encoding));
         } else {
            aBufferedReader = new BufferedReader(new FileReader(file));
         }

         // Look for the start of an HTML title tag.
         while ((line = aBufferedReader.readLine()) != null) {
            i = line.indexOf("<title>");
            if (i<0)
               i = line.indexOf("<TITLE>");
            if (i<0)
               i = line.indexOf("<Title>");
            // If we see it, remember everything after the start tag.
            if (i>=0) {
               title += line.substring(i+7);
               break;
            }
         }  

         // Look for the end of an HTML title tag on the same line as the start tag.
         i = title.indexOf("</title>");
         if (i<0)
            i = title.indexOf("</TITLE>");
         if (i<0)
            i = title.indexOf("</Title>");
         // If we see it, return everything after the start tag and before the end tag.
         if (i>=0) {
            title = title.substring(0,i);
            if (title.length()==0)
               return null;
            return title;
         }

         // If we don't see the end tag on the same line as the start tag, keep
         // appending text to our title until we do see the end tag.
         while ((line = aBufferedReader.readLine()) != null) {
            i = line.indexOf("</title>");
            if (i<0)
               i = line.indexOf("</TITLE>");
            if (i<0)
               i = line.indexOf("</Title>");
            if (i>=0) {
               title += line.substring(0,i);
               break;
            }
            title += line;
         }

      } catch (Throwable e) {
         if (s_debug)
            System.err.println("Error in getTitle("+file.getAbsolutePath()+"): "+e.getMessage());
         return null;
      }

      if (title.length()==0)
         return null;

      return title;
   }

   /**
     * Filter the specified <var>file</var> in accordance with the -filetype option.
     * 
     * @param file File to be examined.  
     * @return     <code>=true</code> if <var>file</var> passes the filters;
     *             otherwise, <code>=false</code>.
     */
   private boolean filterFile(File file) {

      String s = file.getName();

      // If the file type is NOT on the list specified by -filetype,
      // the file should be filtered out.
      for (int i=0; i<this.fileTypes.length; i++) {
         if (s.endsWith(this.fileTypes[i]))
            return true;
      }
      return false;
   }


   /**
     * Streams the current DOM document instance for the site outline to a CDF
     * file in rootDir.
     */
   private void printDocument() {

      OutputStreamWriter anOutputStreamWriter = null;

      try {
         FileOutputStream   aFileOutputStream    = new FileOutputStream(this.rootDir+"/"+s_cdfFileName+".cdf");

         // If no output encoding is specified, or the specified output encoding is
         // not supported by XML document structure services, then go with the default
         // Java encoding.
         if (null == this.outEncoding || null == MIME2Java.convert(this.outEncoding)) {
            anOutputStreamWriter = new OutputStreamWriter(aFileOutputStream);
            ((TXDocument) this.doc).setEncoding(MIME2Java.reverse(anOutputStreamWriter.getEncoding()));
         }

         // If a specified output encoding is specified and supported, use it for
         // our output stream.
         else {
            anOutputStreamWriter = new OutputStreamWriter(aFileOutputStream, MIME2Java.convert(this.outEncoding));
            ((TXDocument) this.doc).setEncoding(this.outEncoding);
         }

         // Stream out the DOM document instance in memory to a CDF file in rootDir.
         ((TXDocument) this.doc).printWithFormat(anOutputStreamWriter);
         aFileOutputStream.close();

      } catch (Exception e) {
         e.printStackTrace();
      }

   }


}  // SiteOutliner
